from pathlib import Path
from readme_renderer import rst
from unittest import TestCase
import io

README_TEXT = (Path(__file__).parents[1] / 'README.rst').read_text()


class TestDoc(TestCase):
    def test_rendering(self):
        out = io.StringIO()
        actual = rst.render(README_TEXT, out)
        if actual is None:
            print('Rendering error!')
            print(out.getvalue())
            assert False



import os
import re
import subprocess
import sys

ENV_PREFIX = 'DOKS_'
DEFAULTS = {'vcstype': 'github'}
FILL_INS = (('packagename', 'repo'),)
SUB = {'stderr': subprocess.DEVNULL, 'encoding': 'utf8'}


def remotes(path):
    cmd = 'git', 'remote', '-v'
    cwd = (path if os.path.isdir(path) else os.path.dirname(path)) or None

    try:
        out = subprocess.check_output(cmd, cwd=cwd, **SUB)
    except Exception:
        return {}

    search = re.compile(r'\bgit@(\w+)\.\w+:(\w+)/(\w+)\.git\b').search
    for remote in out.splitlines():
        match = search(remote)
        if match:
            vcsname, user, repo = match.groups()
            return {'vcsname': vcsname, 'user': user, 'repo': repo}

    return {}


def fill_in(variables, subs=FILL_INS):
    for missing, sub in subs:
        if missing not in variables:
            try:
                variables[missing] = variables[sub]
            except KeyError:
                pass


def env_variables():
    e = {k: v for k, v in os.environ.items() if k.startswith(ENV_PREFIX)}
    p = len(ENV_PREFIX)
    return {k[p:].lower(): v for k, v in e.items()}


def substitute(variables, url):
    parts = url.split('/')
    missing_parts = []
    new_parts = []
    for part in parts:
        if part.startswith(':'):
            optional = part.endswith('*')
            part = part[1:-1] if optional else part[1:]
            replacement = variables.get(part.lower())
            if replacement:
                new_parts.append(replacement)
            elif not optional:
                missing_parts.append(part)
        else:
            new_parts.append(part)

    if missing_parts:
        parts = ', '.join(missing_parts)
        print(variables, file=sys.stderr)
        raise ValueError('Missing variables %s in %s' % (parts, url))

    return '/'.join(new_parts)


def default_variables(path):
    v = dict(DEFAULTS)
    v.update(remotes(path))
    v.update(env_variables())
    fill_in(v)
    return v


def substitutor(path):
    dv = default_variables(path)

    return lambda url: substitute(dv, url)



from doks import from_command
from pathlib import Path
from unittest import TestCase

DOKS_HELP = Path(__file__).parent / 'doks-help.txt'
DOKS_OUTPUT = DOKS_HELP.read_text().splitlines()
DOKS_RESULTS = Path(__file__).parent / 'doks-results.rst'

SIMP_HELP = Path(__file__).parent / 'simp-help.txt'
SIMP_OUTPUT = SIMP_HELP.read_text().splitlines()
SIMP_RESULTS = Path(__file__).parent / 'simp-results.rst'

ARGUMENTS = Path(__file__).parent / 'arguments.txt'


class TestArguments(TestCase):
    def test_arguments(self):
        args = DOKS_OUTPUT[-8:]
        assert args[0] == 'optional arguments:'

        actual = list(from_command._argument(args))
        expected = ARGUMENTS.read_text().splitlines() + ['']
        assert actual == expected


class TestSections(TestCase):
    def test_sections(self):
        actual = list(from_command._sections(DOKS_OUTPUT))
        assert actual == EXPECTED_SECTIONS


class TestFromCommand(TestCase):
    def test_doks(self):
        actual = list(from_command._from_command('doks', DOKS_OUTPUT))
        if False:
            DOKS_RESULTS.write_text('\n'.join(actual) + '\n')

        expected = DOKS_RESULTS.read_text().splitlines()
        assert actual == expected

    def test_simp(self):
        actual = list(from_command._from_command('simp', SIMP_OUTPUT))
        SIMP_RESULTS.write_text('\n'.join(actual) + '\n')

        expected = SIMP_RESULTS.read_text().splitlines()
        assert actual == expected


EXPECTED_SECTIONS = [
    [
        'usage: doks [-h] [--auto] [--command] [--verbose] [--window WINDOW]',
        '            [source] [target]',
        '',
    ],
    ['Write a .rst document for a single Python file', ''],
    [
        'positional arguments:',
        '  source                .py file to create documentation for',
        '  target                .rst file to write to. None means stdout',
        '',
    ],
    [
        'optional arguments:',
        '  -h, --help            show this help message and exit',
        '  --auto, -a            '
        'Automatically guess which files to read and write',
        '  --command, -c         '
        'Use command line help from executing source instead',
        '  --verbose, -v         Print more stuff',
        '  --window WINDOW, -w WINDOW',
        '                        '
        'How many lines around an RST error to print (0 means',
        '                        "print everything")',
    ],
]



"""Top level docs"""

__all__ = 'function', 'Class'


def function(a, b, c='hello', **kwargs):
    """This is a function()."""


class Class:
    "This is a class"

    def __init__(self, *args, **kwargs):
        """a constructor"""
        self.args = args
        self.kwargs = kwargs

    # """before public"""
    def public(self, foo, bar, baz):
        """
        A public method.

        ARGUMENTS
          foo
            a foo

          bar
            a bar

          baz
            no idea
        """

    def undocumented(self, foo, bar, baz):
        pass

    def _protected(self, foo, bar, baz):
        """You shouldn't see this."""



from doks.rst import rst
from unittest import TestCase


class TestSectionCharacters(TestCase):
    def test_empty(self):
        lines = []
        assert rst.section_characters(lines)[:5] == '-=~_+'

    def test_longer(self):
        assert rst.section_characters(LINES)[:5] == '=#~-_'

    def test_ticks(self):
        assert rst.fix_ticks(TICKS) == TICKS_FIXED

    def test_simple(self):
        lines = ['hello', '---', '']
        for i in range(1, 4):
            assert rst.section_characters(lines * i)[:5] == '-=~_+'


TICKS = """
Use `pip <https://pypi.org/project/pip>`_ to install `safer` from the command
line: `pip install safer`.

Tested on Python 3.4 - 3.9.  An old Python 2.7 version
is `here <https://github.com/rec/safer/tree/v2.0.5>`_.
"""

TICKS_FIXED = """
Use `pip <https://pypi.org/project/pip>`_ to install ``safer`` from the command
line: ``pip install safer``.

Tested on Python 3.4 - 3.9.  An old Python 2.7 version
is `here <https://github.com/rec/safer/tree/v2.0.5>`_.
"""

LINES = """
First
======

Second
#######

--------

FIRST!
=======

Second!
#########

Third
~~~~~~

""".splitlines()



#!/usr/bin/env python3
"""
📚 doks: Automatic READMEs from code 📚
====================================================================

Reads the comments from your file and puts them into a servicable .rst
file.

Very suitable for single-file Python libraries that want to keep the code
and documentation in sync without issues

USAGE
-------

.. code-block:: bash

    doks my_file.py [README.rst]

"""
from . import from_command
from . import from_file
from .rst import render
from pathlib import Path
import datetime
import safer
import sys

README = 'README.rst'

__all__ = ('doks',)


def doks(
    source=None,
    target=None,
    auto=False,
    command=False,
    force=False,
    window=None,
    verbose=False,
):
    """
    Write documentation for a file or module.
    Returns True if `target` was written, and False if it was unchanged.

    ARGUMENTS
      source
        path to the Python file or module.

      target
        path to the output file or `None`, in which case
        output is printed to stdout

      auto
        If true, automatically guess both source and target files

      command
        If true, use command line help from executing source file

      force
        If true, write .rst documentation even if it is malformed

      verbose
        Emit more print messages
    """
    if auto:
        source = source or _guess_source()
        target = target or README

    elif not source:
        raise ValueError('Source must be set if --auto/-a is not used')

    if window is None:
        window = render.ERROR_WINDOW

    if command:
        lines = from_command.from_command(source)
    else:
        lines = from_file.from_file(source)
    lines = list(lines)

    if lines and lines[-1]:
        lines.append('')
    lines.append(_DOKS_MSG % _timestamp())

    body = '\n'.join(lines) + '\n'
    if not (render.render(body, window) or force):
        raise ValueError(f'The .rst code in {source} is malformed')

    if not target:
        print(body)
        return True

    p = Path(target)
    written = 'rewritten' if p.exists() else 'written'
    if p.exists() and p.read_text().splitlines()[:-1] == lines[:-1]:
        print(f'{target} unchanged', file=sys.stderr)
        return False

    with safer.writer(target) as fp:
        fp.write(body)

    print(f'{target} {written}', file=sys.stderr)
    return True


def _guess_source():
    eponymous = Path(Path().absolute().stem + '.py')
    if eponymous.exists():
        return eponymous

    def accept(p):
        if p.suffix == '.py' and p.name != 'setup.py':
            return not (p.name.startswith('test') or p.name.startswith('.'))

    files = sorted(p for p in Path().iterdir() if accept(p))
    if not files:
        raise ValueError('No Python files to document')
    if len(files) > 1:
        files = ', '.join(str(f) for f in files)
        raise ValueError(f'Too many possible Python files: {files}')
    return files[0]


def _timestamp():
    return datetime.datetime.now().isoformat()


_DOKS_URL = 'https://github.com/rec/doks/'
_DOKS_MSG = f'(automatically generated by `doks <{_DOKS_URL}>`_ on %s)'



from doks import doks
from doks.from_file import from_file
from pathlib import Path
from unittest import TestCase, mock
import os
import tdir

SAMPLE_FILE = os.path.join(os.path.dirname(__file__), 'sample.py')
TIMESTAMP = '2020-05-28T16:26:07.629835'
RESULTS_FILE = Path(__file__).parent / 'test.rst'
EXPECTED = RESULTS_FILE.read_text().splitlines()[:-1]


class TestDoks(TestCase):
    maxDiff = 100000

    @mock.patch('doks.doks._timestamp', return_value=TIMESTAMP)
    def test_full(self, _timestamp):
        actual = list(from_file(SAMPLE_FILE))
        print(*actual, sep='\n')
        assert actual == EXPECTED

    def test_no_change(self):
        with tdir(chdir=False) as td:
            out = td / 'test.rst'
            assert doks.doks(SAMPLE_FILE, out)

            out_lines = out.read_text().splitlines()
            assert out_lines[:-1] == EXPECTED

            assert not doks.doks(SAMPLE_FILE, out)
            assert out_lines == out.read_text().splitlines()

    @tdir('setup.py', 'test.py', 'result.py')
    def test_auto1(self):
        assert doks._guess_source() == Path('result.py')

    @tdir(fancy=('fancy.py', 'test.py', 'foo.py'))
    def test_auto2(self):
        os.chdir('fancy')
        assert doks._guess_source() == Path('fancy.py')

    @tdir(fancy=('funny.py', 'test.py', 'foo.py'))
    def test_auto3(self):
        os.chdir('fancy')
        with self.assertRaises(ValueError) as m:
            doks._guess_source()
        expected = 'Too many possible Python files: foo.py, funny.py'
        assert m.exception.args[0] == expected

    @tdir('test_all.py', 'test_none.py', 'setup.py')
    def test_auto4(self):
        with self.assertRaises(ValueError) as m:
            doks._guess_source()
        assert m.exception.args[0] == 'No Python files to document'



from doks import shields
import unittest


class TestFindShields(unittest.TestCase):
    def test_empty(self):
        expected = [
            ':tenant.myget',
            ':feed/dt/:packageName',
            'MyGet tenant',
            'downloads',
        ]
        actual = shields._find_shield('')
        assert actual == expected

    def test_github1(self):
        actual = shields._find_shield('github')
        expected = ['github', 'license/:user/:repo', 'GitHub', 'license']
        assert actual == expected

    def test_github2(self):
        actual = shields._find_shield('github.last-commit')
        expected = [
            'github',
            'last-commit/:user/:repo',
            'GitHub last commit',
            'activity',
        ]
        assert actual == expected

    def test_github3(self):
        actual = shields._find_shield('github.release')
        expected = [
            'github',
            'v/release/:user/:repo?include_prereleases&sort=semver',
            'GitHub release (latest SemVer including pre-releases)',
            'version',
        ]
        assert actual == expected

    def test_github4(self):
        actual = shields._find_shield('github.languages/top')
        expected = [
            'github',
            'languages/top/:user/:repo',
            'GitHub top language',
            'analysis',
        ]
        assert actual == expected

    def test_travis(self):
        actual = shields._find_shield('travis.org')
        expected = ['travis', ':user/:repo', 'Travis (.org)', 'build']
        assert actual == expected


_VARS = {
    'user': 'rec',
    'repo': 'doks',
    'vcstype': 'github',
    'vcsname': 'github',
    'packagename': 'doks',
}
_STYLE = {'color': 'FF00FF', 'label': 'blah', 'style': 'plastic'}


class TestShieldURL(unittest.TestCase):
    def test_github2(self):
        actual = shields._shield_url('github.last-commit', _VARS)
        expected = 'https://img.shields.io/github/last-commit/rec/doks'
        assert actual == (expected, 'GitHub last commit')

    def test_travis(self):
        actual = shields._shield_url('travis.org', _VARS)
        expected = ('https://img.shields.io/travis/rec/doks', 'Travis (.org)')
        assert actual == expected

    def test_travis2(self):
        actual = shields._shield_url('travis.org', _VARS, _STYLE)[0]
        expected = (
            'https://img.shields.io/travis/rec/doks'
            '?color=FF00FF&label=blah&style=plastic'
        )
        assert actual == expected


class TestAddShields(unittest.TestCase):
    def test_add(self):
        actual = '\n'.join(shields._add_shields(LINES.splitlines(), _VARS))
        print(actual)
        assert actual == EXPECTED


LINES = """
-------------------------------
✏️safer: a safer file opener ✏️
-------------------------------

.. doks-shields::

   travis.org codecov github.release pypi.pyversions github.top/languages
   codefactor pypi.l github.code-size

No more partial writes or corruption! For file streams, sockets or
any callable.
"""

EXPECTED = """
-------------------------------
✏️safer: a safer file opener ✏️
-------------------------------

|doks_0| |doks_1| |doks_2| |doks_3| |doks_4| |doks_5| |doks_6| |doks_7|

.. |doks_0| image:: https://img.shields.io/travis/rec/doks
   :alt: Travis (.org)
   :target: https://img.shields.io/travis/rec/doks

.. |doks_1| image:: https://img.shields.io/codecov/c/github/rec/doks
   :alt: Codecov
   :target: https://img.shields.io/codecov/c/github/rec/doks

.. |doks_2| image:: https://img.shields.io/github/v/release/rec/doks
   :alt: GitHub release (latest SemVer including pre-releases)
   :target: https://img.shields.io/github/v/release/rec/doks

.. |doks_3| image:: https://img.shields.io/pypi/pyversions/doks
   :alt: PyPI - Python Version
   :target: https://img.shields.io/pypi/pyversions/doks

.. |doks_4| image:: https://img.shields.io/github/languages/top/rec/doks
   :alt: GitHub top language
   :target: https://img.shields.io/github/languages/top/rec/doks

.. |doks_5| image:: https://img.shields.io/codefactor/grade/github/rec/doks
   :alt: CodeFactor Grade
   :target: https://img.shields.io/codefactor/grade/github/rec/doks

.. |doks_6| image:: https://img.shields.io/pypi/l/doks
   :alt: PyPI - License
   :target: https://img.shields.io/pypi/l/doks

.. |doks_7| image:: https://img.shields.io/github/languages/code-size/rec/doks
   :alt: GitHub code size in bytes
   :target: https://img.shields.io/github/languages/code-size/rec/doks

No more partial writes or corruption! For file streams, sockets or
any callable."""



from . import doks
from pathlib import Path
import argparse
import sys
import toml
import traceback

PYP = Path('pyproject.toml')


def main():
    p = argparse.ArgumentParser(description=_DESCRIPTION)

    p.add_argument('source', default=None, nargs='?', help=_SOURCE_HELP)
    p.add_argument('target', default=None, nargs='?', help=_TARGET_HELP)
    p.add_argument('--auto', '-a', action='store_true', help=_AUTO_HELP)
    p.add_argument('--command', '-c', action='store_true', help=_COMMAND_HELP)
    p.add_argument('--force', '-f', action='store_true', help=_FORCE_HELP)
    p.add_argument('--verbose', '-v', action='store_true', help=_VERBOSE_HELP)
    p.add_argument('--window', '-w', help=_WINDOW_HELP)

    args = vars(p.parse_args())

    if PYP.exists():
        project = toml.load(PYP.open())
        project = project.get('tool', {}).get('doks', {})
    else:
        project = {}

    bad_project = set(project) - set(args)
    if bad_project:
        bp = sorted(bad_project)
        print('Do not understand values in', PYP, ':', *bp, file=sys.stderr)
        project = {k: v for k, v in project.items() if k in args}

    for k, v in args.items():
        if not (k in project or v in (False, None)):
            project[k] = v

    try:
        doks.doks(**project)
    except Exception as e:
        print(e, file=sys.stderr)
        if args.get('verbose'):
            traceback.print_exc()
        sys.exit(-1)


_DESCRIPTION = """Write a .rst document for a single Python file"""
_SOURCE_HELP = '.py file to create documentation for'
_TARGET_HELP = '.rst file to write to.  None means stdout'
_AUTO_HELP = """Automatically guess which files to read and write"""
_COMMAND_HELP = """Use command line help from executing source instead"""
_FORCE_HELP = """Write .rst documentation even if it is malformed"""
_VERBOSE_HELP = """Print more stuff"""
_WINDOW_HELP = """How many lines around an RST error to print
(0 means "print everything")"""


if __name__ == '__main__':
    main()



from . import shields
from .rst import describe
from .rst import rst
import impall
import inspect


def from_file(path):
    m = impall.import_file(str(path))
    doks = getattr(m, '_DOKS', {})
    try:
        module = m._xmod_wrapped
    except AttributeError:
        module = m
        extends = None
    else:
        extends = m._xmod_extension

    module_doc = rst.fix_ticks(inspect.getdoc(module) or '')

    lines = module_doc.splitlines()
    sections = rst.section_characters(lines)

    yield from shields.add_shields(lines, path)
    yield ''
    yield from rst.header('API', sections[1])

    items = getattr(module, '__all__', None)
    if not items:
        raise ValueError('No items in module')

    if extends:
        d = path.with_suffix('')
        yield from describe.describe(d, extends, sections, False, doks)

    for vpath, value, is_member in _children(module, items, module.__name__):
        if value is not extends:
            yield from describe.describe(
                vpath, value, sections, is_member, doks
            )


def _children(parent, names, module_path, is_member=False):
    for name in names:
        if not name.startswith('_') or name.startswith('__'):
            value = getattr(parent, name)
            if callable(value):
                path = '%s.%s' % (module_path, name)
                yield path, value, is_member

                if isinstance(value, type):
                    yield from _children(value, vars(value), path, True)



from .rst import rst
from pathlib import Path
import os
import re
import subprocess

USAGE = 'usage: '


def from_command(path):
    env = dict(os.environ, PYTHONPATH=str(Path('.').absolute()))
    try:
        output = subprocess.check_output((path, '-h'), env=env)
    except Exception:
        output = subprocess.check_output(('python', path, '-h'), env=env)

    lines = output.decode().splitlines()
    yield from _from_command(Path(path).name, lines)


def _arguments(title, *args):
    name = title.split()[0]
    assert name in ('positional', 'optional')
    is_pos = name == 'positional'

    yield from _header('%s arguments' % name.capitalize())
    for i, arg in enumerate(args):
        a = arg.lstrip()
        if not a:
            yield a
            continue
        spaces = len(arg) - len(a)

        if spaces > 4:
            yield '  ' + a
        elif is_pos:
            if i:
                yield ''
            arg, doc = a.split(maxsplit=1)
            yield '``{}``'.format(arg)
            yield '  ' + doc
        else:
            if i:
                yield ''
            flags = []
            doc = []
            in_doc = False

            for part in re.split(r'(\S+)', a):
                if len(part) > 1 and part.isspace():
                    in_doc = True
                else:
                    (doc if in_doc else flags).append(part)

            flags = ''.join(flags).split(',')
            yield ', '.join('``{}``'.format(f.strip()) for f in flags)

            if doc:
                doc[0] = doc[0].capitalize()
                yield '  ' + ''.join(doc)


def _from_command(name, lines):
    usage, desc, pos, opt, *rest = _sections(lines)

    yield from _header(name, 0)

    assert usage[0].startswith(USAGE)
    usage[0] = ' ' * len(USAGE) + usage[0][len(USAGE) :]

    yield from _header('Usage')
    yield from rst.code_block('bash')
    yield from usage

    yield from _header('Description')
    yield from desc

    yield from _arguments(*pos)
    yield from _arguments(*opt)

    if rest:
        (epilog,) = rest
        yield ''
        yield from _header('Comments')
        yield from epilog


def _argument(lines):
    section = []
    for line in lines:
        indent = len(line) - len(line.lstrip())
        if section and indent and indent <= 4:
            yield from section + ['']
            section = []
        section.append(line)

    if section:
        yield from section + ['']


def _sections(lines):
    section = []
    for line in lines:
        if line.strip():
            section.append(line)
        elif section:
            section.append(line)
            yield section
            section = []

    if section:
        yield section


def _header(line, index=1):
    yield from rst.header(line, rst.SECTIONS[index])



from bs4 import BeautifulSoup
import requests
import safer
import yaml

SHIELDS_URL = 'https://shields.io'
FILE = 'shields.yml'
INTERMEDIATE_FILE = 'shields-inter.yml'


def get_soup(url=''):
    page = requests.get(SHIELDS_URL + url)
    return BeautifulSoup(page.text, features='html.parser')


def get_categories():
    soup = get_soup()
    links = [a.get('href') for a in soup.find_all('a')]
    return sorted(i for i in links if i.startswith('/category'))


def get_shield_template(category):
    for row in get_soup(category).find('table').find('tbody').find_all('tr'):
        yield row.find('th').text.rstrip(':'), row.find('code').text


def scrape_all_shields():
    results = {}

    for category in get_categories():
        name = category.split('/')[-1]
        results[name] = result = {}
        for name, template in get_shield_template(category):
            result[name] = template

    return results


def get_all_shields():
    with open(INTERMEDIATE_FILE) as fp:
        return yaml.safe_load(fp)


def re_key(shields):
    result = {}

    for category, items in shields.items():
        for alt, item in items.items():
            blank, root, *rest = item.split('/', maxsplit=2)
            assert not blank, blank
            url = rest[0] if rest else ''
            result.setdefault(root, []).append([alt, url, category])

    return result


def print_re_key(shields):
    with safer.printer(FILE) as print:
        first = True
        for root, items in sorted(shields.items()):
            if first:
                first = False
            else:
                print()
            print('%s:' % root)
            for alt, url, category in sorted(items):
                print("- ['%s', '%s', '%s']" % (url, alt, category))


def main():
    scraped = scrape_all_shields()
    print_re_key(re_key(scraped))


if __name__ == '__main__':
    main()



from doks import variables
from unittest import TestCase, mock
import os

RESULTS = """\
origin	git@ONE.com:TWO/THREE.git (fetch)
origin	git@github.com:rec/doks.git (push)

# comment

"""

EXPECTED = {
    'bar': 'bing',
    'foo': 'foo',
    'packagename': 'THREE',
    'repo': 'THREE',
    'user': 'TWO',
    'vcsname': 'ONE',
    'vcstype': 'github',
}


@mock.patch.dict(os.environ, {'DOKS_FOO': 'foo', 'DOKS_BAR': 'bing'})
@mock.patch('subprocess.check_output', return_value=RESULTS)
class TestVariables(TestCase):
    def test_simple(self, check_output):
        actual = variables.default_variables('.')
        assert EXPECTED == actual

    def test_substitute(self, check_output):
        sub = variables.substitutor('.')
        actual = sub('tests/:user/:repo')
        assert actual == 'tests/TWO/THREE'

    def test_missing(self, check_output):
        sub = variables.substitutor('.')
        with self.assertRaises(ValueError) as m:
            sub('tests/:user/:repong')
        expected = 'Missing variables repong in tests/:user/:repong'
        actual = m.exception.args[0]
        assert actual == expected

    def test_optional(self, check_output):
        sub = variables.substitutor('.')
        actual = sub('grade/:vcsType/:user/:repo/:branch*')
        expected = 'grade/github/TWO/THREE'
        assert actual == expected



from .variables import default_variables, substitute
import os
import shlex
import yaml

__all__ = ('add_shields',)

_SHIELD_DATA = {}
_URL_ROOT = 'https://img.shields.io'
FILE = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'shields.yml')
PREFIX = '.. doks-shield'


def add_shields(lines, path):
    yield from _add_shields(lines, default_variables(path))


def _add_shields(lines, variables):
    shield_lines = []
    in_shield = False
    for line in lines:
        if line.startswith(PREFIX):
            in_shield = True
        elif not in_shield:
            yield line
        elif not line.strip():
            continue
        elif line.startswith(' '):
            shield_lines.append(line)
        else:
            in_shield = False
            shields = shlex.split(' '.join(shield_lines))
            yield ' '.join('|doks_%d|' % i for i in range(len(shields)))
            yield ''
            for i, shield in enumerate(shields):
                url, alt = _parse_shield(shield, variables)
                target = url  # FIX
                yield '.. |doks_%d| image:: %s' % (i, url)
                yield '   :alt: ' + alt
                yield '   :target: ' + target
                yield ''

            yield line


def _shield_data():
    if not _SHIELD_DATA:
        with open(FILE) as fp:
            _SHIELD_DATA.update(yaml.safe_load(fp))

    return _SHIELD_DATA


def _find(shield_key):
    key, *rest = shield_key.lower().split('.')

    for source, items in _shield_data().items():
        if not source.lower().startswith(key):
            continue

        if not rest:
            return [source] + items[0]
        k1, *r1 = rest

        for url, alt, category in items:
            result = [source, url, alt, category]
            keys = k1.split('/')

            if k1 and not all(k in url.split('/') for k in keys):
                continue

            if not r1:
                return result
            k2, *r2 = r1
            if k2 not in alt:
                continue

            if not r2:
                return result
            (k3,) = r2
            if category.startswith(k3):
                return result


def _find_shield(key):
    shield = _find(key) or _find(key.replace('.', '..', 1))
    if not shield:
        raise ValueError('Bad key ' + key)
    return shield


def _shield_url(key, variables, style=None):
    source, url, alt, category = _find_shield(key)
    url = url.split('?', maxsplit=1)[0]
    url = substitute(variables, url)

    base_url = '/'.join([_URL_ROOT, source, url])
    if style:
        s = ('%s=%s' % (k, v) for k, v in sorted(style.items()))
        base_url += '?' + '&'.join(s)

    return base_url, alt


def _parse_shield(key, variables):
    style = None
    if '{' in key:
        sd = yaml.safe_load(key)
        key = sd.pop('key')
        variables = dict(variables, **sd.pop('variables', {}))
        style = sd.pop('style', None)
        if sd:
            raise ValueError('Unknown variables %s' % sd)

    return _shield_url(key, variables, style)
